import java.awt.image.*;
import javax.imageio.*;
import java.io.*;
import java.awt.event.*;

public class Main {

  public static final int SQRT_SAMPLES = 16;

  public static final int IMAGE_SCALE = 6;
  public static final int WIDTH = 1920 / IMAGE_SCALE;
  public static final int HEIGHT = 1080 / IMAGE_SCALE;
  public static final double VIRTUAL_SCREEN_WIDTH = 100;
  public static final int FRAMES = 30;

  public static final double[] EYE = { 2, 100, -2 };
  public static final double[] LOOK = { 1000, 77, -1000 };
  public static final double[] LIGHT = { -564, 686, 147 };
  public static final double[] LIGHT_COLOR = { 1, 1, 1 };
  public static final double LIGHT_RADIUS = 0.0;
  public static final double[] AMBIENT_COLOR = { 1, 1, 1 };
  public static double DISTANCE_TO_VIRTUAL_SCREEN = 50.0;
  public static double GROUND_SQUARE_SIZE = 107.0;

  public static final String IMAGE_TYPE = "png";
  public static final boolean RENDER_IN_WINDOW = true;

  public static final double GAMMA = 2.2;

  public static final int SAMPLES = SQRT_SAMPLES * SQRT_SAMPLES;
  public static final double INVERSE_SQRT_SAMPLES = 1.0 / SQRT_SAMPLES;
  public static final double INVERSE_SAMPLES = 1.0 / SAMPLES;
  public static final double INVERSE_GAMMA = 1.0 / GAMMA;
  public static final double HALF_WIDTH = WIDTH / 2.0;
  public static final double HALF_HEIGHT = HEIGHT / 2.0;
  public static final double VIRTUAL_SCREEN_RATIO
      = VIRTUAL_SCREEN_WIDTH / WIDTH;
  public static double INVERSE_GROUND_SQUARE_SIZE = 1.0 / GROUND_SQUARE_SIZE;

  public static final boolean AMBIENT_OCCLUSION = true;
  public static final double MAX_OCCLUSION_DISTANCE = 100.0;
  public static final double RADIANCE_SCALE = 1.0;
  public static final double EPSILON = 1E-6;
  public static final int MAX_DEPTH = 10;
  public static final double MIN_COLOR_INTENSITY = 1.0 / 256.0;

  public static final long SECOND_MILLIS = 1000L;
  public static final long MINUTE_MILLIS = 60 * SECOND_MILLIS;
  public static final long HOUR_MILLIS = 60 * MINUTE_MILLIS;

  public static final double JUGGLE_X0 = -182;
  public static final double JUGGLE_X1 = -108;
  public static final double JUGGLE_Y0 = 88;
  public static final double JUGGLE_H_Y = 184;

  public static final double JUGGLE_H_VX = (JUGGLE_X0 - JUGGLE_X1) / 60.0;
  public static final double JUGGLE_L_VX = (JUGGLE_X1 - JUGGLE_X0) / 30.0;

  public static final double JUGGLE_H_H = JUGGLE_H_Y - JUGGLE_Y0;
  public static final double JUGGLE_H_VY = 4.0 * JUGGLE_H_H / 60.0;
  public static final double JUGGLE_G = JUGGLE_H_VY * JUGGLE_H_VY
      / (2.0 * JUGGLE_H_H);

  public static final double JUGGLE_L_VY = 0.5 * JUGGLE_G * 30.0;

  private RenderFrame renderFrame;
  private BufferedImage image;
  private int runningCount;
  private int rowIndex;

  private volatile double[][][] blendPixels;
  private int keyCode;
  private boolean keyPressed;

  // 96, 110, -106
  private volatile double _X = 155;
  private volatile double _Y = 0;
  private volatile double _Z = -164;

  public void launch() throws Throwable {
    
    //loadImage("amiga03.png");

    image = new BufferedImage(WIDTH, HEIGHT, BufferedImage.TYPE_INT_RGB);
    if (RENDER_IN_WINDOW) {
      renderFrame = new RenderFrame(image);
      addKeyListener();
    }
    
    int processors = Runtime.getRuntime().availableProcessors();    

    for(int frameIndex = 0; frameIndex < FRAMES; frameIndex++) {

    //while(true) {

      if (RENDER_IN_WINDOW) {
        renderFrame.setTitle("Amiga Juggler [" + frameIndex + "]");
      }

      resetRowIndex();
      updateRunningCount(processors);
      for(int i = 0; i < processors; i++) {
        final int _frameIndex = frameIndex;
        new Thread(Integer.toString(i)) {
          @Override
          public void run() {
            render(_frameIndex);
            updateRunningCount(-1);
          }
        }.start();
      }

      synchronized(this) {
        while(runningCount != 0) {
          wait();
        }
      }

      saveImage("juggler" + ((frameIndex < 10) ? "0" : "") + frameIndex);

      /*synchronized(this) {
        while(!keyPressed) {          
          wait();
        }

        keyPressed = false;
        switch(keyCode) {
          case KeyEvent.VK_LEFT:
            _X -= 1;
            break;
          case KeyEvent.VK_RIGHT:
            _X += 1;
            break;
          case KeyEvent.VK_UP:
            _Y += 1;
            break;
          case KeyEvent.VK_DOWN:
            _Y -= 1;
            break;
          case KeyEvent.VK_D:
            _Z += 1;
            break;
          case KeyEvent.VK_E:
            _Z -= 1;
            break;
        }

        System.out.format("%f %f %f%n", _X, _Y, _Z);
      }*/
    }
  }

  private void addKeyListener() {
    renderFrame.addKeyListener(new KeyAdapter() {
      @Override
      public void keyPressed(KeyEvent e) {
        synchronized(Main.this) {
          keyCode = e.getKeyCode();
          keyPressed = true;
          Main.this.notifyAll();
        }
      }
    });
  }
 
  private void render(int frameIndex) {

    IObject[] scene = createScene();

    double[] u = new double[3];
    double[] v = new double[3];
    double[] w = new double[3];
    double[] c = new double[3];
    double[] p = new double[3];
    double[] d = new double[3];
    double[] l = new double[3];
    double[] l2 = new double[3];
    double[] light = new double[3];
    double[] r = new double[3];
    double[] ar = new double[3];
    double[] f = new double[3];
    double[] o = new double[3];
    double[][] temps = new double[16][3];
    Intersection intersection = new Intersection();
    Intersection bestIntersection = new Intersection();

    Vec.constructUnitVector(w, EYE, LOOK);
    Vec.ray(c, EYE, w, -DISTANCE_TO_VIRTUAL_SCREEN);
    Vec.onb(u, v, w);

    RandomRays randomRays = new RandomRays();
    RandomDoubles randomDoubles = new RandomDoubles();
    RandomTimes randomTimes = new RandomTimes();
    int[] pixels = new int[WIDTH];
    double[] pixel = new double[3];

    while(true) {

      int y = getNextRowIndex();
      if (y >= HEIGHT) {
        return;
      }

      for(int x = 0; x < WIDTH; x++) {

        Vec.assign(pixel, 0, 0, 0);

        double[] times = randomTimes.random();
        int timesIndex = 0;

        for(int i = 0; i < SQRT_SAMPLES; i++) {

          double b = y + INVERSE_SQRT_SAMPLES * i;
          if (SQRT_SAMPLES == 1) {
            b += 0.5;
          } else {
            b += randomDoubles.random();
          }

          b = VIRTUAL_SCREEN_RATIO * (HALF_HEIGHT - b);

          for(int j = 0; j < SQRT_SAMPLES; j++) {
            double a = x + INVERSE_SQRT_SAMPLES * j;
            if (SQRT_SAMPLES == 1) {
              a += 0.5;
            } else {
              a += randomDoubles.random();
            }

            a = VIRTUAL_SCREEN_RATIO * (a - HALF_WIDTH);

            Vec.map(p, c, u, v, a, b);
            Vec.constructUnitVector(d, p, EYE);

            Vec.assign(o, EYE);
            Vec.assign(f, RADIANCE_SCALE, RADIANCE_SCALE, RADIANCE_SCALE);

            updateScene(frameIndex, times[timesIndex++], scene, temps);

            reflectionLoop: for(int m = 0; m < MAX_DEPTH; m++) {

              boolean hit = false;
              bestIntersection.time = Double.POSITIVE_INFINITY;
              for(int k = scene.length - 1; k >= 0; k--) {
                IObject object = scene[k];
                if (object.intersect(o, d, true, Double.POSITIVE_INFINITY,
                    temps, intersection)) {
                  if (intersection.time < bestIntersection.time) {
                    hit = true;
                    bestIntersection.time = intersection.time;
                    Vec.assign(bestIntersection.normal, intersection.normal);
                    Vec.assign(bestIntersection.hit, intersection.hit);
                    bestIntersection.material = intersection.material;
                  }
                }
              }

              if (hit) {

                if (Vec.dot(bestIntersection.normal, d) >= 0) {
                  Vec.negate(bestIntersection.normal);
                }

                if (bestIntersection.material.ambientWeight > 0) {
                  double ambientPercent = 1.0;

                  if (AMBIENT_OCCLUSION
                      && bestIntersection.material.ambientOcclusionPercent > 0) {

                    randomRays.randomRay(ar);
                    if (Vec.dot(ar, bestIntersection.normal) < 0) {
                      Vec.negate(ar);
                    }

                    for(int k = scene.length - 1; k >= 0; k--) {
                      IObject object = scene[k];
                      if (object.intersect(bestIntersection.hit, ar, false,
                          MAX_OCCLUSION_DISTANCE, temps, intersection)) {
                        ambientPercent = 1.0
                            - bestIntersection.material.ambientOcclusionPercent;
                        break;
                      }
                    }
                  }

                  for(int k = 0; k < 3; k++) {
                    pixel[k] += bestIntersection.material.ambientWeight
                        * bestIntersection.material.diffuseColor[k]
                        * f[k]
                        * AMBIENT_COLOR[k]
                        * ambientPercent;
                  }
                }

                Vec.assign(light, LIGHT);
                if (LIGHT_RADIUS > 0) {
                  Vec.subtract(l2, bestIntersection.hit, light);
                  randomRays.randomRay(l);
                  if (Vec.dot(l, l2) < 0) {
                    Vec.negate(l);
                  }
                  Vec.ray(light, LIGHT, l, LIGHT_RADIUS);
                } 
                Vec.constructUnitVector(l, light, bestIntersection.hit);

                double nDotl = Vec.dot(l, bestIntersection.normal);
                if (nDotl > 0) {

                  double maxTime = Vec.distance(bestIntersection.hit, light);

                  boolean illuminated = true;
                  for(int k = scene.length - 1; k >= 0; k--) {
                    IObject object = scene[k];
                    if (object.intersect(bestIntersection.hit, l, false,
                        maxTime, temps, intersection)) {
                      illuminated = false;
                      break;
                    }
                  }

                  if (illuminated) {

                    if (bestIntersection.material.diffuseWeight > 0) {
                      for(int k = 0; k < 3; k++) {
                        pixel[k] += bestIntersection.material.diffuseWeight
                            * bestIntersection.material.diffuseColor[k]
                            * f[k]
                            * LIGHT_COLOR[k]
                            * nDotl;
                      }
                    }

                    if (bestIntersection.material.specularWeight > 0) {

                      Vec.scale(r, bestIntersection.normal, 2.0 * nDotl);
                      Vec.subtract(r, l);
                      double rDotMd = Vec.dotNegative(r, d);
                      if (rDotMd > 0) {
                        for(int k = 0; k < 3; k++) {
                          pixel[k] += bestIntersection.material.specularWeight
                              * Math.pow(rDotMd,
                                  bestIntersection.material.shininess)
                              * f[k]
                              * LIGHT_COLOR[k]
                              * nDotl
                              * bestIntersection.material.highlightColor[k];
                        }
                      }
                    }
                  }
                }

                if (bestIntersection.material.reflectionWeight > 0) {
                  boolean aboveMinColorIntensity = false;
                  for(int k = 0; k < 3; k++) {
                    f[k] *= bestIntersection.material.reflectionWeight
                        * bestIntersection.material.reflectionColor[k];
                    if (f[k] >= MIN_COLOR_INTENSITY) {                      
                      aboveMinColorIntensity = true;
                    }
                  }
                  if (!aboveMinColorIntensity) {
                    break reflectionLoop;
                  }

                  Vec.ray(d, bestIntersection.normal,
                      -2 * Vec.dot(bestIntersection.normal, d));
                  Vec.assign(o, bestIntersection.hit);
                } else {
                  break reflectionLoop;
                }

              } else {
                break reflectionLoop;
              }
            }
          }
        }

        int value = 0;
        for(int i = 0; i < 3; i++) {

          // TODO remove this later
          //pixel[i] = 0.5 * pixel[i] + 0.5 * blendPixels[y][x][i];

          int intensity = (int)Math.round(255
              * Math.pow(pixel[i] * INVERSE_SAMPLES, INVERSE_GAMMA));
          if (intensity < 0) {
            intensity = 0;
          } else if (intensity > 255) {
            intensity = 255;
          }
          value <<= 8;
          value |= intensity;
        }

        pixels[x] = value;
      }

      rowCompleted(y, pixels);
    }
  }

  private IObject[] createScene() {

    /*return new IObject[] {
      new Ground(25, Materials.YELLOW_MATTE, Materials.GREEN_MATTE), // 0
      new Sphere(0, 0, 0, 1E6, Materials.BLUE_MATTE),                // 1
      
      new Sphere(0, 0, 0, 10, Materials.RED_PLASTIC), // 2
      new Sphere(0, 0, 0, 10, Materials.RED_PLASTIC), // 3
      new Sphere(0, 0, 0, 10, Materials.RED_PLASTIC), // 4
    };*/

    IObject[] scene = new IObject[2 + 3];
    scene[0] = new Ground(Materials.YELLOW_MATTE, Materials.GREEN_MATTE);
    scene[1] = new Sphere(0, 0, 0, 1E6, Materials.BLUE_MATTE);
    for(int i = 2; i <= 4; i++) {
      scene[i] = new Sphere(110, 0, 0, 14, Materials.MIRROR);
    }

    // juggling balls 2 -- 4

    return scene;

    //return new IObject[] {
    //  new Ground(Materials.YELLOW_MATTE, Materials.GREEN_MATTE),
    //  new Sphere(0, 0, 0, 1E6, Materials.BLUE_MATTE),

      // head
      //new Sphere(151, 155, -151, 14, Materials.RED_PLASTIC),

      // juggling balls
      //new Sphere(110, 88, -108, 14, Materials.RED_PLASTIC),
      //new Sphere(110, 88, -182, 14, Materials.RED_PLASTIC),
      //new Sphere(110, 184, -145, 14, Materials.RED_PLASTIC),

      // right foot
      //new Sphere(139, 0, -164, 5, Materials.RED_PLASTIC),

      // left foot
      //new Sphere(159, 0, -133, 5, Materials.RED_PLASTIC),

      // neck
      //new Sphere(151, 140, -151, 5, Materials.RED_PLASTIC),

      // chest
      //new Sphere(151, 117, -151, 20, Materials.RED_PLASTIC),

      // left shoulder
      //new Sphere(151, 130, -132, 5, Materials.RED_PLASTIC),

      // left elbo
      //new Sphere(151, 88, -113, 5, Materials.RED_PLASTIC),

      // left hand
      //new Sphere(110, 72, -108, 3, Materials.RED_PLASTIC),

      // hips
      //new Sphere(151, 85, -151, 16, Materials.RED_PLASTIC),

      // left hip
      //new Sphere(151, 76, -135, 5, Materials.RED_PLASTIC),
    //};
  }

  private void updateScene(
      int frameIndex, double t, IObject[] scene, double[][] temps) {

    double T = (30 + frameIndex + t);
    Sphere sphere = (Sphere)scene[3];
    sphere.center[2] = JUGGLE_X1 + JUGGLE_H_VX * T;
    sphere.center[1] = JUGGLE_Y0 + (JUGGLE_H_VY - 0.5 * JUGGLE_G * T) * T;

    T = frameIndex + t;
    sphere = (Sphere)scene[4];
    sphere.center[2] = JUGGLE_X1 + JUGGLE_H_VX * T;
    sphere.center[1] = JUGGLE_Y0 + (JUGGLE_H_VY - 0.5 * JUGGLE_G * T) * T;

    sphere = (Sphere)scene[2];
    sphere.center[2] = JUGGLE_X0 + JUGGLE_L_VX * T;
    sphere.center[1] = JUGGLE_Y0 + (JUGGLE_L_VY - 0.5 * JUGGLE_G * T) * T;

    /*double[] w = temps[5];
    w[0] = 0;
    w[1] = 0;
    w[2] = 1;

    double[] p = temps[6];
    p[0] = -75;
    p[1] = 50;
    p[2] = 0;

    double angle = 2.0 * Math.PI * (frameIndex + t) / 30.0;
    double[] q = temps[7];
    q[0] = 75 + 5 * Math.cos(angle);
    q[1] = 50 + 5 * Math.sin(angle);
    q[2] = 0;

    updateAppendage(scene, 2, p, q, w, 80, 80, 8, 8, temps);*/
  }

  private void updateAppendage(
      IObject[] scene, int sceneIndex,
      double[] p, double[] q, double[] w, double A, double B,
      int countA, int countB, double[][] temps) {

    double[] U = temps[0];
    double[] V = temps[1];
    double[] W = temps[2];
    double[] j = temps[3];
    double[] d = temps[4];

    Vec.subtract(V, q, p);
    double D = Vec.magnitude(V);
    double inverseD = 1.0 / D;
    Vec.scale(V, inverseD);

    Vec.normalize(W, w);
    Vec.cross(U, V, W);

    double A2 = A * A;

    double y = 0.5 * inverseD * (A2 - B * B + D * D);
    double x = Math.sqrt(A2 - y * y);

    Vec.map(j, p, U, V, x, y);

    Vec.subtract(d, j, p);
    Vec.scale(d, 1.0 / countA);
    for(int i = 0; i <= countA; i++) {
      double[] center = ((Sphere)scene[sceneIndex + i]).center;
      Vec.ray(center, p, d, i);
    }

    Vec.subtract(d, j, q);
    Vec.scale(d, 1.0 / countB);
    for(int i = 0; i < countB; i++) {
      double[] center = ((Sphere)scene[countA + 1 + sceneIndex + i]).center;
      Vec.ray(center, q, d, i);      
    }
  }

  private synchronized void updateRunningCount(int dx) {
    runningCount += dx;
    if (runningCount == 0) {
      notifyAll();
    }
  }

  private synchronized void resetRowIndex() {
    rowIndex = 0;
  }

  private synchronized int getNextRowIndex() {
    return rowIndex++;
  }

  private synchronized void rowCompleted(int rowIndex, int[] pixels) {
    image.setRGB(0, rowIndex, WIDTH, 1, pixels, 0, WIDTH);
    if (RENDER_IN_WINDOW) {
      renderFrame.imageUpdated();
    }
  }

  private void loadImage(String name) throws Throwable {
    BufferedImage blendImage = ImageIO.read(new File(name));
    blendPixels = new double[blendImage.getHeight()][blendImage.getWidth()][3];
    int[] row = new int[blendImage.getWidth()];
    for(int i = 0; i < blendImage.getHeight(); i++) {
      blendImage.getRGB(0, i, blendImage.getWidth(), 1, row, 0,
          blendImage.getWidth());
      for(int j = 0; j < blendImage.getWidth(); j++) {
        for(int k = 2; k >= 0; k--) {
          int intensity = row[j] & 0xFF;
          row[j] >>= 8;
          blendPixels[i][j][k] = Math.pow(intensity / 255.0, GAMMA);
        }
      }
    }
  }

  private void saveImage(String name) throws Throwable {
    ImageIO.write(image, IMAGE_TYPE, new File(name + "." + IMAGE_TYPE));
  }

  public static void main(String... args) throws Throwable {

    long startTime = System.currentTimeMillis();

    Main main = new Main();
    main.launch();

    long interval = System.currentTimeMillis() - startTime;
    long hours = interval / HOUR_MILLIS;
    interval %= HOUR_MILLIS;
    long minutes = interval / MINUTE_MILLIS;
    interval %= MINUTE_MILLIS;
    long seconds = interval / SECOND_MILLIS;
    interval %= SECOND_MILLIS;
    System.out.format("%d hour%s, %d minute%s, %d second%s, %d millisecond%s%n",
        hours, hours == 1 ? "" : "s",
        minutes, minutes == 1 ? "" : "s",
        seconds, seconds == 1 ? "" : "s",
        interval, interval == 1 ? "" : "s");
  }

}
